/*
 * Copyright (C) 2012 - Juan Ferrer Toribio
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "vn-shelves.h"
#include "vn-allocator.c"

#define $(id) (vn_form_get (VN_FORM (self), id))

G_DEFINE_TYPE (VnShelves, vn_shelves, VN_TYPE_FORM);

//+++++++++++++++++++++++++++++++++++++++++++++++++++ Private

#define LINE_WIDTH 10
#define PAGE_MARGIN 40
#define HEADER_HEIGHT 80
#define TITLE_SIZE 40
#define BOX_TEXT_SIZE 35
#define BOX_MARGIN 6

static
GdkRGBA colors[] = {
	 {0.96, 0.86, 0.86, 1} // Red
	,{0.86, 0.96, 0.86, 1} // Green
	,{0.86, 0.86, 0.96, 1} // Blue
	,{0.96, 0.96, 0.76, 1} // Yellow
};

//-------------------------- Drawing methods

void
vn_shelves_draw_box (VnAllocator * allocator, VnShelves * self, Item * i, gint amount)
{
	if (i->box_width == 0 || i->box_height == 0)
		return;

	VnShelf shelf = self->shelf;
	cairo_t * cr = self->cr;

	gint x = self->shelf_x + allocator->tray_x + shelf.hspacing;
	gint y = self->shelf_y - allocator->tray_y
		- shelf.first_tray_elevation - shelf.tray_density
		- allocator->current_tray * (shelf.tray_height + shelf.tray_density);

	cairo_rectangle (cr, x, y, i->box_width, -i->box_height);
	
	static gint n_item = -1;
	
	if (amount == 0)
		n_item++;

	cairo_save (cr);
	gint n_color = n_item % G_N_ELEMENTS (colors);
	gdk_cairo_set_source_rgba (cr, &colors[n_color]);
	cairo_fill_preserve (cr);
	cairo_restore (cr);

	cairo_stroke (cr);

	if (amount == 0 || allocator->first_shelf_box)
	{
		PangoRectangle extents;
	
		PangoFontDescription * font = pango_font_description_new ();
		pango_font_description_set_family (font, "Sans");
		pango_font_description_set_absolute_size (font, BOX_TEXT_SIZE * PANGO_SCALE);

		PangoLayout * layout = pango_cairo_create_layout (cr);
		pango_layout_set_font_description (layout, font);
		pango_layout_set_width (layout, (i->box_width - BOX_MARGIN * 2) * PANGO_SCALE);
		pango_layout_set_height (layout, (i->box_height - BOX_MARGIN * 2) * PANGO_SCALE);
		pango_layout_set_ellipsize (layout, PANGO_ELLIPSIZE_START);
		pango_layout_set_text (layout, i->name, -1);
		pango_layout_set_alignment (layout, PANGO_ALIGN_CENTER);
		pango_layout_get_extents (layout, NULL, &extents);

		cairo_move_to (cr
			,x + BOX_MARGIN
			,y - BOX_MARGIN - i->box_height / 2 - (extents.height / PANGO_SCALE) / 2
		);
		pango_cairo_show_layout (cr, layout);

		g_object_unref (layout);
		pango_font_description_free (font);
	}
	
	cairo_stroke (cr);
}

void
vn_shelves_draw_shelf (VnAllocator * allocator, VnShelves * self)
{
	gint y;
	VnShelf shelf = self->shelf;

	cairo_surface_t * surface = cairo_surface_create_similar (
		cairo_get_target (self->context_cr),
		CAIRO_CONTENT_COLOR_ALPHA,
		self->width,
		self->height
	);
	self->pages = g_list_append (self->pages, surface);

	if (self->cr)
		cairo_destroy (self->cr);

	cairo_t * cr = self->cr = cairo_create (surface);	
	cairo_select_font_face (cr, "Sans"
		,CAIRO_FONT_SLANT_NORMAL
		,CAIRO_FONT_WEIGHT_NORMAL
	);

	cairo_scale (cr, self->scale, self->scale);
	cairo_set_line_width (cr, LINE_WIDTH);

	cairo_save (cr);
	cairo_scale (cr, 1 / self->scale, 1 / self->scale);

	// Drawing the title

	cairo_text_extents_t extents;
	
	const gchar * title = gvn_param_get_string ($("title"));

	cairo_set_font_size (cr, TITLE_SIZE);
	cairo_text_extents (cr, title, &extents);
	y = PAGE_MARGIN + extents.height;

	cairo_move_to (cr, PAGE_MARGIN, y);
	cairo_show_text (cr, title);

	char * shelf_number = g_strdup_printf ("%d", allocator->current_shelf + 1);
	cairo_text_extents (cr, shelf_number, &extents);
	cairo_move_to (cr, self->width - PAGE_MARGIN - extents.width, y);
	cairo_show_text (cr, shelf_number);
	g_free (shelf_number);

	cairo_restore (cr);

	// Calculating the shelf bottom left coordinates

	gint shelf_x = (self->width / 2) / self->scale;
	gint shelf_y = (self->height - PAGE_MARGIN) / self->scale;

	shelf_x -= shelf.width / 2;

	self->shelf_x = shelf_x;
	self->shelf_y = shelf_y;

	// Drawing the shelf

	cairo_move_to (cr, shelf_x, shelf_y);
	cairo_rel_line_to (cr, 0, -shelf.height);

	cairo_move_to (cr, shelf_x + shelf.width, shelf_y);
	cairo_rel_line_to (cr, 0, -shelf.height);

	// Drawing trays

	int last_tray_y = shelf.first_tray_elevation;

	if (shelf.tray_height > 0)
	while (last_tray_y + shelf.tray_density < shelf.height)
	{
		y = shelf_y - last_tray_y;

		cairo_move_to (cr, shelf_x, y);
		cairo_rel_line_to (cr, shelf.width, 0);

		cairo_move_to (cr, shelf_x, y - shelf.tray_density);
		cairo_rel_line_to (cr, shelf.width, 0);

		last_tray_y += shelf.tray_height + shelf.tray_density;
	}

	cairo_stroke (cr);
}

void
vn_shelves_on_begin_print (GtkPrintOperation * op, GtkPrintContext * context, VnShelves * self)
{
	DbIterator * iter;

	self->pages = NULL;
	self->context_cr = gtk_print_context_get_cairo_context (context);
	
	// XXX: Shelf dimensions

	iter = $("iter-shelves");
	VnShelf shelf = {
		 db_iterator_get_int (iter, "width")
		,db_iterator_get_int (iter, "height")
		,db_iterator_get_int (iter, "max_height")
		,db_iterator_get_int (iter, "tray_height")
		,db_iterator_get_int (iter, "first_tray_elevation")
		,db_iterator_get_int (iter, "tray_density")
		,db_iterator_get_int (iter, "vspacing")
		,db_iterator_get_int (iter, "hspacing")
	};
	self->shelf = shelf;

	// XXX: Items list
		
	Item * i;
	iter = $("iter-items");
	gint nitems = db_iterator_get_nrows (iter);
	
	Item * items = g_new (Item, nitems + 1);
	items[nitems].name = NULL;

	db_iterator_move_iter (iter, NULL);

	for (i = items; db_iterator_move_next (iter); i++)
	{
		i->name = g_strdup_printf ("%s x%d"
			,db_iterator_get_string (iter, "Article")
			,db_iterator_get_int (iter, "Medida")
		);
		i->box_width = db_iterator_get_double (iter, "x") * 10;
		i->box_height = db_iterator_get_double (iter, "z") * 10;
		i->amount = db_iterator_get_double (iter, "etiquetas");
	}
	
	// Initializing

	self->scale = 0.0;
	self->width = gtk_print_context_get_width (context) /* 210 */;
	self->height = gtk_print_context_get_height (context) /* 297 */;
	
	// Calculating the shelf scale

	{
		gint w = self->width - PAGE_MARGIN * 2;
		gint h = self->height - PAGE_MARGIN * 2 - HEADER_HEIGHT;

		if (h < 0)
			h = 0;
		if (w < 0)
			w = 0;

		self->scale = (gdouble) h / shelf.max_height;
	
		if (shelf.width * self->scale > w)
			self->scale = (gdouble) w / shelf.width;

		self->scale = self->scale;
	}

	// Generating the pages
	
	gint max_amount = gvn_param_get_double ($("max-amount"));

	VnAllocator alloc;
	alloc.items = items;
	alloc.max_amount = max_amount;
	alloc.shelf_func = (VnAllocatorShelfFunc) vn_shelves_draw_shelf;
	alloc.box_func = (VnAllocatorBoxFunc) vn_shelves_draw_box;
	alloc.user_data = self;

	alloc.ntrays = ceil (
		(gdouble) (shelf.height - shelf.first_tray_elevation) /
		(shelf.tray_height + shelf.tray_density)
	);
	alloc.tray_width = shelf.width - shelf.hspacing * 2;
	alloc.tray_height = shelf.tray_height - shelf.vspacing;
	alloc.top_tray_height = shelf.max_height - shelf.vspacing
		- shelf.first_tray_elevation - (alloc.ntrays - 1) * shelf.tray_height;

	gint npages = vn_allocator_run (&alloc);
	
	gtk_print_operation_set_n_pages (op, npages);
	
	// XXX: Freeing items list
	
	for (i = items; i->name; i++)
		g_free (i->name);

	g_free (items);
}

void
vn_shelves_on_end_print (GtkPrintOperation * op, GtkPrintContext * context, VnShelves * self)
{
	g_list_free_full (self->pages, (GDestroyNotify) cairo_surface_destroy);
	cairo_destroy (self->cr);
	self->cr = NULL;
}

void
vn_shelves_on_draw_page (GtkPrintOperation * op, GtkPrintContext * context, gint page_num, VnShelves * self)
{
	cairo_t * cr = gtk_print_context_get_cairo_context (context);
	cairo_surface_t * surface = g_list_nth_data (self->pages, page_num);
	cairo_set_source_surface (cr, surface, 0, 0);
	cairo_paint (cr);
}

void
vn_shelves_show_message (VnShelves * self, const gchar * message)
{
	GtkWindow * parent = GTK_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (self)));
	GtkWidget * dialog = gtk_message_dialog_new (parent
		,GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT
		,GTK_MESSAGE_INFO
		,GTK_BUTTONS_CLOSE
		,message
	);
	gtk_dialog_run (GTK_DIALOG (dialog));
	gtk_widget_destroy (GTK_WIDGET (dialog));
}

void
vn_shelves_on_items_changed (DbModel * model, DbModelStatus status, VnShelves * self)
{
	gtk_widget_set_sensitive ($("buttonbox"),
		status != DB_MODEL_STATUS_LOADING);

	if (status != DB_MODEL_STATUS_READY || self->action == -1)
		return;

	if (db_model_get_nrows (model) == 0)
	{
		vn_shelves_show_message (self, _("There are no items to allocate"));
		return;
	}	

	GtkPrintOperation * op = gtk_print_operation_new ();
	
	if (self->print_settings)
		gtk_print_operation_set_print_settings (op, self->print_settings);

	gtk_print_operation_set_job_name (op, _("Shelves"));
	gtk_print_operation_set_embed_page_setup (op, TRUE);
	g_object_connect (op
		,"signal::begin-print", vn_shelves_on_begin_print, self
		,"signal::end-print", vn_shelves_on_end_print, self
		,"signal::draw-page", vn_shelves_on_draw_page, self
		,NULL
	);

	GError * err = NULL;
	GtkWindow * parent = GTK_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (self)));
	GtkPrintOperationResult result =
		gtk_print_operation_run (op, self->action, parent, &err);
	
	if (result == GTK_PRINT_OPERATION_RESULT_APPLY)
	{
		self->print_settings =
			g_object_ref (gtk_print_operation_get_print_settings (op));
	}
	else if (result == GTK_PRINT_OPERATION_RESULT_ERROR && err)
	{
		vn_shelves_show_message (self, err->message);
		g_error_free (err);
	}
}

void
vn_shelves_preview_clicked (GtkButton * button, VnShelves * self)
{
	self->action = GTK_PRINT_OPERATION_ACTION_PREVIEW;
	db_model_refresh ($("items"));
}

void
vn_shelves_print_clicked (GtkButton * button, VnShelves * self)
{
	self->action = GTK_PRINT_OPERATION_ACTION_PRINT_DIALOG;
	db_model_refresh ($("items"));
}

//-------------------------- Form

void
vn_shelves_on_edit_clicked (GtkButton * button, VnShelves * self)
{
	GtkWidget * dialog = $("shelves-dialog");
	gtk_window_set_transient_for (GTK_WINDOW (dialog),
		GTK_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (self))));
	gtk_widget_show_all (dialog);
	gtk_dialog_run (GTK_DIALOG (dialog));
	gtk_widget_hide (dialog);
}

void
vn_shelves_on_close_clicked (GtkButton * button, VnShelves * self)
{
	gtk_dialog_response ($("shelves-dialog"), 1);
}

static void
vn_shelves_open (VnShelves * self, gpointer user_data)
{
	// XXX: Default field values
	
	self->action = -1;

	gvn_param_set_string ($("title"), "Anthuriums");
	gvn_param_set_string ($("filter"), "a.Article LIKE \\'Ant %\\'");
	gvn_param_set_int ($("wh"), 1);
	gvn_param_set_int ($("shelf"), 1);
	
	GValue date = G_VALUE_INIT;
	g_value_init (&date, G_TYPE_DATE_TIME);
	g_value_take_boxed (&date, g_date_time_new_now_local ());
	gvn_param_set_value ($("date"), &date);
	g_value_unset (&date);
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++ Class

static void
vn_shelves_init (VnShelves * self)
{
	self->print_settings = NULL;
}

static void
vn_shelves_finalize (VnShelves * self)
{
	g_clear_object (&self->print_settings);
	G_OBJECT_CLASS (vn_shelves_parent_class)->finalize (G_OBJECT (self));
}

static void
vn_shelves_class_init (VnShelvesClass * k)
{
	G_OBJECT_CLASS (k)->finalize = (GObjectFinalizeFunc) vn_shelves_finalize;
	VN_FORM_CLASS (k)->open = (VnFormOpenFunc) vn_shelves_open;
}
